unit MainUnit;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ExtCtrls, StdCtrls;

const
  AM_GOTDATA = WM_USER + $99;
  AM_DONE = WM_USER + $98;

type
  TMainForm = class(TForm)
    DeviceGroup: TGroupBox;
    DeviceCombo: TComboBox;
    DeviceDetectButton: TButton;
    ConfigPanel: TPanel;
    DurationEdit: TEdit;
    Label1: TLabel;
    GoPanel: TPanel;
    TimestampLabel: TLabel;
    ADCLabel0: TLabel;
    ADCLabel1: TLabel;
    LogMemo: TMemo;
    QuadLabel: TLabel;
    LogFNButton: TButton;
    ArmButton: TButton;
    StopButton: TButton;
    LiveTimestampLabel: TLabel;
    procedure FormCreate(Sender: TObject);
    procedure ArmButtonClick(Sender: TObject);
    procedure StopButtonClick(Sender: TObject);
    procedure DeviceDetectButtonClick(Sender: TObject);
    procedure DeviceComboChange(Sender: TObject);
    procedure DurationEditChange(Sender: TObject);
    procedure DurationEditExit(Sender: TObject);
    procedure LogFNButtonClick(Sender: TObject);
  protected
    DeviceIndex: LongWord;
    bProgrammaticChange: Boolean;

    procedure DurationEditValidate;
    procedure AMGotData(var Msg: TMessage); message AM_GOTDATA;
    procedure AMDone(var Msg: TMessage); message AM_DONE;
    { Private declarations }
  public
    { Public declarations }
  end;

var
  MainForm: TMainForm;

implementation

{$R *.DFM}

uses
  Math,
  Registry, ShellAPI,
  AIOUSB,
  LogThreads;



type
  TUSBAI162APoint = packed record
    ADC0, ADC1: Word;
    DITS: LongWord;
  end;
  TPUSBAI162APoint = ^TUSBAI162APoint;

  TPRefLink = ^TRefLink;
  TRefLink = record
    pNext: TPRefLink;
    BufSize: LongWord;
    pBuf: Pointer;
  end;

  TParseThread = class(TThread)
  public
    constructor Create(NewLogThread: TLogThread); virtual;
    destructor Destroy; override;

    procedure AddBuf(pBuf: Pointer; BufSize: LongWord); virtual;
  protected
    pFirstRef, pLastRef: TPRefLink;
    hRefSem, hRefMutex: THandle;
    LogThread: TLogThread;

    procedure Execute; override;
    procedure ParseBuf(pDataPoint: TPUSBAI162APoint; WholePoints: Integer); virtual;
  end;

const
  MinDuration = 1/10;
  MaxDuration = $07FFFFFF div 1000000;

var
  LogThread: TLogThread;
  ParseThread: TParseThread;
  LeftoverData: AnsiString;
  Config: array[0..1] of Byte;
  LastTimestamp, DisplayTimestamp: LongWord;
  LastV: array[0..1] of Double;
  LastDIOData, QuadI: Integer;
  GoMode: (gmArmed, gmTriggered, gmDone);
  TriggerTimestamp: LongWord;
  usDuration: LongWord;
  bPostedDone: Boolean;

procedure ADCallback(pBuf: PWord; BufSize, Flags, Context: LongWord); cdecl; forward;

function IntToBinN(I, NBits: Integer): String;
var
  iCh: Integer;

begin
  SetString(Result, nil, NBits);
  for iCh := NBits downto 1 do begin
    Result[iCh] := Char(Byte('0') + (I and 1));
    I := I shr 1;
  end;
end;

function MultiStrCat(const Strs: array of String): String;
var
  iStr, iCh, L: Integer;

begin
  L := 0;
  for iStr := 0 to High(Strs) do
    Inc(L, Length(Strs[iStr]))
  ;
  SetString(Result, nil, L);

  iCh := 1;
  for iStr := 0 to High(Strs) do begin
    L := Length(Strs[iStr]);
    Move(Strs[iStr][1], Result[iCh], L);
    Inc(iCh, L);
  end;
end;

procedure LoadDefaultFont(IntoFont: TFont);
var
  FontSize: Integer;
  NCM: TNonClientMetrics;

begin
  NCM.cbSize := SizeOf(NCM);
  if SystemParametersInfo(SPI_GETNONCLIENTMETRICS, SizeOf(NCM), @NCM, 0) then begin
    FontSize := Round(-3/4 * NCM.lfMessageFont.lfHeight);
    IntoFont.Name := NCM.lfMessageFont.lfFaceName;
  end
  else begin
    FontSize := Abs(IntoFont.Size);
    IntoFont.Name := 'Tahoma';
  end;
  IntoFont.PixelsPerInch := 72;
  IntoFont.Size := FontSize;
end;

procedure EnabledWithChildren(WinControl: TWinControl; NewEnabled: Boolean);
var
  I: Integer;

begin
  WinControl.Enabled := NewEnabled;
  for I := 0 to WinControl.ControlCount-1 do
    WinControl.Controls[I].Enabled := NewEnabled
  ;
end;

function VoltsFromCounts(Ct: Word; RangeCode: Byte): Double;
begin
  Result := 1/65536 * Ct;
  if (RangeCode and 1) <> 0 then Result := Result * 2 - 1;
  if (RangeCode and 2) = 0 then Result := Result * 2;
  if (RangeCode and 4) = 0 then Result := Result * 5;
end;

{ TParseThread }

procedure TParseThread.AddBuf(pBuf: Pointer; BufSize: LongWord);
var
  pRefLink: TPRefLink;

begin
  New(pRefLink);
  pRefLink.pNext := nil;
  pRefLink.BufSize := BufSize;
  pRefLink.pBuf := PByte(LocalAlloc(LMEM_FIXED, BufSize));
  Move(pBuf^, pRefLink.pBuf^, BufSize);

  WaitForSingleObject(hRefMutex, INFINITE);
    //Add it to the linked list.
    if pLastRef <> nil then
      pLastRef.pNext := pRefLink
    else
      pFirstRef := pRefLink
    ;
    pLastRef := pRefLink;
  ReleaseMutex(hRefMutex);

  ReleaseSemaphore(hRefSem, 1, nil);
end;

constructor TParseThread.Create(NewLogThread: TLogThread);
begin
  inherited Create(True);

  LogThread := NewLogThread;
  hRefSem := CreateSemaphore(nil, 0, $FFFFFF, nil);
  hRefMutex := CreateMutex(nil, False, nil);

  Resume;
end;

destructor TParseThread.Destroy;
begin
  //Send an extra "kill release" to the semaphore, without a corresponding link in the list.
  ReleaseSemaphore(hRefSem, 1, nil);

  inherited;

  CloseHandle(hRefSem);
  CloseHandle(hRefMutex);
end;

procedure TParseThread.Execute;
var
  pRefLink: TPRefLink;

begin
  LastDIOData := -1;
  QuadI := 0;
  LastTimestamp := 0;
  DisplayTimestamp := 0;
  LastV[0] := 0;
  LastV[1] := 0;
  GoMode := gmArmed;

  SetThreadPriority(GetCurrentThread, THREAD_PRIORITY_BELOW_NORMAL);
  try
    repeat
      WaitForSingleObject(hRefSem, INFINITE);

      WaitForSingleObject(hRefMutex, INFINITE);
        pRefLink := pFirstRef;
        if pRefLink = nil then begin
        end
        else if pRefLink.pNext = nil then begin
          pFirstRef := nil;
          pLastRef := nil;
        end
        else begin
          pFirstRef := pRefLink.pNext;
        end;
      ReleaseMutex(hRefMutex);

      if pRefLink = nil then Break; //It was a "kill release".

      if GoMode <> gmDone then
        ParseBuf(pRefLink.pBuf, pRefLink.BufSize div SizeOf(TUSBAI162APoint))
      ;

      LocalFree(THandle(pRefLink.pBuf));
      Dispose(pRefLink);
    until False;
  except
    on E: Exception do MessageBox(0, PChar(E.Message), PChar(String(E.ClassName)), MB_ICONEXCLAMATION);
  end;
end;

procedure TParseThread.ParseBuf(pDataPoint: TPUSBAI162APoint; WholePoints: Integer);
const
  BitChar: array[0..1] of Char = ('0', '1');

var
  iPoint, iLine: Integer;
  LineStr: array of String;
  DIOData: Byte;
  usTimestamp: LongWord;
  ADC0V, ADC1V: Double;
  ErrStr: String;

begin
  if GoMode = gmTriggered then
    SetLength(LineStr, WholePoints)
  else
    SetLength(LineStr, 0)
  ;
  ErrStr := '';
  iLine := 0;
  for iPoint := 0 to WholePoints-1 do begin
    //Parse the DIO bits and timestamp, for triggering.
    DIOData := (pDataPoint.DITS shr 1) and $F;
    usTimestamp := pDataPoint.DITS shr 5;

    if GoMode = gmArmed then begin
      //A change on DIO 2 is a trigger.
      if LastDIOData >= 0 then begin
        if ((DIOData xor LastDIOData) and $4) <> 0 then begin
          SetLength(LineStr, WholePoints - iPoint);
          TriggerTimestamp := usTimestamp;
          GoMode := gmTriggered;
        end;
      end;
    end;

    if GoMode = gmTriggered then begin
      if usTimestamp = LastTimestamp+1 then
      //Check the timestamp for the end of the duration.
      if ((usTimestamp - TriggerTimestamp) and $07FFFFFF) > usDuration then begin
        SetLength(LineStr, iLine);

        //LogThread.WriteStr(MultiStrCat(['Ended at ', IntToStr(usTimestamp), '-', IntToStr(TriggerTimestamp), '. ', IntToStr(iLine), ' scans to be posted.'#13#10]));

        if not bPostedDone then begin
          //Notify the main thread we're done.
          PostMessage(MainForm.Handle, AM_DONE, 0, 0);
        end;

        GoMode := gmDone;
      end;
    end;

    if GoMode = gmTriggered then begin
      //Parse the analog data.
      ADC0V := VoltsFromCounts(pDataPoint.ADC0, Config[0]);
      ADC1V := VoltsFromCounts(pDataPoint.ADC1, Config[1]);

      //Parse the quadrature.
      case (DIOData xor LastDIOData) and $3 of
        1: begin //Whenever A changes, movement is positive if the new states differ, and negative if the new states match.
          if ((DIOData xor (DIOData shr 1)) and 1) <> 0 then
            Inc(QuadI)
          else
            Dec(QuadI)
          ;
        end;
        2: begin //Whenever B changes, movement is negative if the new states differ, and positive if the new states match.
          if ((DIOData xor (DIOData shr 1)) and 1) <> 0 then
            Dec(QuadI)
          else
            Inc(QuadI)
          ;
        end;
        3: begin
          ErrStr := ',"AB Simultaneous Change"';
        end;
      end;

      //Convert to string.
      LineStr[iLine] := MultiStrCat
      ([
        IntToStr(usTimestamp - TriggerTimestamp), ',',
        FloatToStr(ADC0V), ',',
        FloatToStr(ADC1V), ',',
        IntToStr(QuadI), ',',
        BitChar[(DIOData      ) and 1], ',',
        BitChar[(DIOData shr 1) and 1], ',',
        BitChar[(DIOData shr 2) and 1], ',',
        BitChar[(DIOData shr 3) and 1], ErrStr, #13#10
      ]);
      Inc(iLine);
      ErrStr := '';

      //Ready for display.
      LastTimestamp := usTimestamp;
      LastV[0] := ADC0V;
      LastV[1] := ADC1V;
    end;

    //For triggering and quadrature.
    LastDIOData := DIOData;

    //Advance to the next point. (Or after the last point, to the leftover data if any.)
    Inc(pDataPoint);
  end;

  if Length(LineStr) > 0 then begin
    LogThread.WriteStr(MultiStrCat(LineStr));

    //Notify the display that we have data ready for it.
    PostMessage(MainForm.Handle, AM_GOTDATA, 0, 0);
  end;
end;

{ TMainForm }

procedure TMainForm.FormCreate(Sender: TObject);
begin
  IsMultiThread := True;

  LoadDefaultFont(Font);

  DeviceDetectButton.Click;
end;

procedure TMainForm.DeviceDetectButtonClick(Sender: TObject);
var
  iDeviceIndex, MaskDevices, Status, PID, DeviceNameSize: LongWord;
  DeviceName: AnsiString;

begin
  DeviceIndex := diNone;
  DeviceCombo.Items.Clear;

  for iDeviceIndex := 0 to 31 do AIOUSB_CloseDevice(iDeviceIndex);
  MaskDevices := GetDevices;
  for iDeviceIndex := 0 to 31 do
    if (MaskDevices and (1 shl iDeviceIndex)) <> 0 then begin
      DeviceNameSize := 256;
      SetString(DeviceName, nil, DeviceNameSize);
      Status := QueryDeviceInfo(iDeviceIndex, @PID, @DeviceNameSize, @DeviceName[1], nil, nil);
      if Status <> ERROR_SUCCESS then Continue;

      if PID <> $8180 then Continue; //Not what we're looking for.

      SetLength(DeviceName, DeviceNameSize);

      DeviceCombo.Items.Add('[' + IntToStr(iDeviceIndex) + '] ' + DeviceName + '(' + IntToHex(PID, 4) + ')');
    end
  ;
  if DeviceCombo.Items.Count <> 0 then begin
    DeviceCombo.ItemIndex := 0;
    DeviceComboChange(DeviceCombo);
  end;
end;

procedure TMainForm.DurationEditChange(Sender: TObject);
var
  sNewDuration: Double;
  Code: Integer;

begin
  if bProgrammaticChange then Exit;

  Val(DurationEdit.Text, sNewDuration, Code);
  if (Code = 0) and (sNewDuration >= MinDuration) and (sNewDuration <= MaxDuration) then begin
    usDuration := Round(sNewDuration * 1000000);
    DurationEdit.Color := clWindow;
    DurationEdit.Font.Color := clWindowText;
  end
  else begin
    DurationEdit.Color := $0000FF;
    DurationEdit.Font.Color := $FFFFFF;
  end;
end;

procedure TMainForm.DurationEditExit(Sender: TObject);
begin
  DurationEditValidate;
end;

procedure TMainForm.DurationEditValidate;
var
  sNewDuration: Double;
  Code, iCh: Integer;
  DurStr, NewDurStr: String;

begin
  DurStr := DurationEdit.Text;
  Val(DurStr, sNewDuration, Code);
  if Code <> 0 then begin
    sNewDuration := 0;
    for iCh := 1 to Length(DurStr) do
      if DurStr[iCh] in ['0'..'9'] then
        sNewDuration := sNewDuration * 10 + (Ord(DurStr[iCh]) - Ord('0'))
    ;
  end;
  sNewDuration := Min(MaxDuration, Max(MinDuration, sNewDuration));

  NewDurStr := FloatToStr(sNewDuration);
  if (usDuration <> sNewDuration * 1000000) or (DurStr <> NewDurStr) or (DurationEdit.Color <> clWindow) then begin
    bProgrammaticChange := True;
      usDuration := Round(sNewDuration * 1000000);
      DurationEdit.Text := NewDurStr;
      DurationEdit.Color := clWindow;
      DurationEdit.Font.Color := clWindowText;
    bProgrammaticChange := False;
  end;
end;

procedure TMainForm.ArmButtonClick(Sender: TObject);
var
  Status: LongWord;
  LogFN: String;
  ST: TSystemTime;
  ConfigSize: LongWord;
  BufSize, BufCount: LongWord;

begin
  DurationEditValidate;

  LogMemo.Lines.Add('Starting');

  GetLocalTime(ST);
  with ST do LogFN := Format('Log %.4d%.2d%.2d_%.2d%.2d%.2d.csv', [ wYear, wMonth, wDay, wHour, wMinute, wSecond ]);
  LogThread := TLogThread.Create(LogFN);
  LogThread.WriteStr('"s","A/D 0","A/D 1","Quad","D0","D1","D2","D3"'#13#10);
  LogFNButton.Caption := LogFN;
  LogFNButton.Visible := True;

  Config[0] := 6; //0-1V `@ What's used for 4-20mA?
  Config[1] := 6; //0-1V
  ConfigSize := Length(Config);
  Status := ADC_SetConfig(DeviceIndex, @Config[0], ConfigSize);
  if Status <> ERROR_SUCCESS then begin
    LogMemo.Lines.Add('Error ' + IntToStr(Status) + ' from ADC_SetConfig.');
    LogThread.Free;
    Exit;
  end;

  AIOUSB_ClearFIFO(DeviceIndex, 0);
  bPostedDone := False;
  LeftoverData := '';
  ParseThread := TParseThread.Create(LogThread);

  LiveTimestampLabel.Caption := '';
  TimestampLabel.Caption := 'Armed';
  ADCLabel0.Caption := '';
  ADCLabel1.Caption := '';
  QuadLabel.Caption := '';

  BufSize := 1024*1024;
  BufCount := 64;
  Status := ADC_BulkContinuousCallbackStart(DeviceIndex, BufSize, BufCount, 0, @ADCallback);
  if Status <> ERROR_SUCCESS then begin
    LogMemo.Lines.Add('Error ' + IntToStr(Status) + ' from ADC_BulkContinuousCallbackStart.');
    ParseThread.Free;
    LogThread.Free;
    Exit;
  end;

  LogMemo.Lines.Add('Started.');

  StopButton.Enabled := True;
  ArmButton.Enabled := False;
  EnabledWithChildren(DeviceGroup, False);
  EnabledWithChildren(ConfigPanel, False);

  //Drop the priority of the UI thread by 1.
  SetThreadPriority(GetCurrentThread, THREAD_PRIORITY_BELOW_NORMAL);
end;

procedure ADCallback(pBuf: PWord; BufSize, Flags, Context: LongWord); cdecl;
var
  pDataPoint: TPUSBAI162APoint;
  ThisData: AnsiString;
  WholePoints: Integer;
  oL: LongWord;

begin
  if BufSize = 0 then Exit; //We don't care about Flags.

  oL := Length(LeftoverData);
  if oL = 0 then begin
    //Just get the new data.
    SetString(ThisData, PAnsiChar(pBuf), BufSize);
  end
  else begin
    //Append the new data onto the previous leftover data.
    SetString(ThisData, nil, oL + BufSize);
    Move(LeftoverData[1], ThisData[1], oL);
    Move(pBuf^, ThisData[1 + oL], BufSize);
  end;

  WholePoints := Length(ThisData) div SizeOf(pDataPoint^);

  pDataPoint := @ThisData[1];

  if WholePoints > 0 then begin
    //There's whole points of data, so parse it.
    ParseThread.AddBuf(pDataPoint, WholePoints * SizeOf(pDataPoint^));

    //Advance to the last point, and fetch the timestamp for display.
    Inc(pDataPoint, WholePoints-1);
    DisplayTimestamp := pDataPoint.DITS shr 5;

    //If we've gone over the duration, then the parse thread is falling behind. Notify the main thread we're done.
    if (not bPostedDone) and (GoMode = gmTriggered) and (((DisplayTimestamp - TriggerTimestamp) and $07FFFFFF) > usDuration + 2048) then begin
      PostMessage(MainForm.Handle, AM_DONE, 0, 0);

      bPostedDone := True;
    end;

    //Advance to after the last point, to the leftover data if any.
    Inc(pDataPoint);
  end;

  //Update the leftover data for next time.
  SetString(LeftoverData, PAnsiChar(pDataPoint), Length(ThisData) - WholePoints*SizeOf(pDataPoint^));

  //Notify the display that we have data ready for it.
  PostMessage(MainForm.Handle, AM_GOTDATA, 0, 0);
end;

procedure TMainForm.AMGotData(var Msg: TMessage);
begin
  Msg.Result := ERROR_SUCCESS;

  case GoMode of
    gmArmed: begin
      LiveTimestampLabel.Caption := 'Timestamp: ' + IntToStr(DisplayTimestamp) + 's';
    end;
    gmTriggered: begin
      LiveTimestampLabel.Caption := 'Live Timestamp: ' + IntToStr(DisplayTimestamp) + 's';
      TimestampLabel.Caption := 'Parse Timestamp: ' + IntToStr(LastTimestamp) + 's';
      ADCLabel0.Caption := 'ADC 0: ' + FloatToStrF(LastV[0], ffFixed, 16, 3) + 'V';
      ADCLabel1.Caption := 'ADC 1: ' + FloatToStrF(LastV[1], ffFixed, 16, 3) + 'V';
      QuadLabel.Caption := 'Quad: ' + IntToStr(QuadI);
    end;
    gmDone: begin
      LiveTimestampLabel.Caption := '';
      TimestampLabel.Caption := 'Done.';
      ADCLabel0.Caption := '';
      ADCLabel1.Caption := '';
      QuadLabel.Caption := '';
    end;
  end;
end;

procedure TMainForm.AMDone(var Msg: TMessage);
begin
  Msg.Result := ERROR_SUCCESS;

  if StopButton.Enabled then StopButton.Click;
end;

procedure TMainForm.StopButtonClick(Sender: TObject);
var
  Status, IOStatus: LongWord;
  Msg: TMessage;

begin
  //Restore the priority of the UI thread.
  SetThreadPriority(GetCurrentThread, THREAD_PRIORITY_NORMAL);

  LogMemo.Lines.Add('Stopping');
  Status := ADC_BulkContinuousEnd(DeviceIndex, @IOStatus);
  if Status <> ERROR_SUCCESS then
    LogMemo.Lines.Add('Error ' + IntToStr(Status) + ' from ADC_BulkContinuousEnd.')
  else if IOStatus <> ERROR_SUCCESS then
    LogMemo.Lines.Add('I/O error ' + IntToStr(IOStatus) + ' occurred.')
  else
    LogMemo.Lines.Add('Stopped.')
  ;

  ParseThread.Free;
  LogThread.Free;

  //`@ LogMemo.Lines.Add('Logged up to ' + IntToStr(LastTimestamp) + 's.');

  GoMode := gmDone;
  Msg.Msg := AM_GOTDATA;
  Msg.WParam := 0;
  Msg.LParam := 0;
  AMGotData(Msg);

  ArmButton.Enabled := True;
  StopButton.Enabled := False;
  EnabledWithChildren(DeviceGroup, True);
  EnabledWithChildren(ConfigPanel, True);
end;

procedure TMainForm.DeviceComboChange(Sender: TObject);
var
  I: Integer;
  DevStr: String;

begin
  DeviceIndex := diNone;

  I := DeviceCombo.ItemIndex;
  if I < 0 then Exit;

  DevStr := DeviceCombo.Items[I];
  DevStr := Copy(DevStr, 2, MAXINT); //Cut off [.
  I := Pos(']', DevStr);
  DevStr := Copy(DevStr, 1, I-1); //Cut off ] and the rest.

  DeviceIndex := StrToInt(DevStr);
end;

procedure TMainForm.LogFNButtonClick(Sender: TObject);
var
  FN, Dir: String;

begin
  FN := LogFNButton.Caption;
  Dir := GetCurrentDir;
  ShellExecute(Handle, nil, PChar(FN), nil, PChar(Dir), SW_SHOWNORMAL);
  FN := '';
  Dir := '';
end;

end.
